// PitchforkPanel.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"
#include "resource.h"
#include "commctrl.h"
#include <stdlib.h>

extern HMODULE hModule;
const int dllVersion = 110;
const int MT4_REFRESH = 33324;
const int MAX_LINES = 1000;
const int MAX_NAME_LEN = 79;
const int MAX_COMMANDS = 100;
const int MAX_WINDOWS = 200;
const int MAX_COLORS = 100;
const int LAST_LINE = 4;
const int CMD_SCHIFF = 0;
const int CMD_UNSCHIFF = 1;
const int CMD_WL = 2;
const int CMD_DELETE_WL = 3;
const int CMD_COPYSCHIFF  = 4;
const int CMD_COPYANDREWS = 5;
const int CMD_PARALLEL = 6;

const int controlMap[][2] = {
	{150, IDC_WL1A},
	{200, IDC_WL2A},
	{250, IDC_WL3A},
	{300, IDC_WL4A},
	{-50, IDC_WL1B},
	{-100, IDC_WL2B},
	{-150, IDC_WL3B},
	{-200, IDC_WL4B},
	{75, IDC_QL1A},
	{125, IDC_QL2A},
	{175, IDC_QL3A},
	{225, IDC_QL4A},
	{275, IDC_QL5A},
	{25, IDC_QL1B},
	{-25, IDC_QL2B},
	{-75, IDC_QL3B},
	{-125, IDC_QL4B},
	{-175, IDC_QL5B},
	{0, -1}};

struct Command
{
	int id;
	char objectName[MAX_NAME_LEN + 1];
	int param;
};

struct ColorData
{
	COLORREF color;
	int imageUp;
	int imageDn;
};

// Shared data segment. An instance of the dll runs for each chart so common
// stuff must be shared. C++ objects can't be used here.

#pragma data_seg(push, "SHARED")
const char registryKey[] = "Software\\PitchforkPanel";
const char windowKey[] = "WindowRect";
const char snapKey[] = "SnapToPivot";
HKEY hKey = NULL;
HWND hDlg = NULL;
HWND hMtWnd = NULL;
int timerTicks = 0;
char lastForkName[MAX_NAME_LEN + 1] = "";
char lineNames[MAX_LINES][MAX_NAME_LEN + 1] = {{0}};
int lineCount = 0;
Command commandList[MAX_COMMANDS] = {{0, 0, 0}};
int commandCount = 0;
HWND hwndList[MAX_WINDOWS] = {NULL};
int hwndCount = 0;
HIMAGELIST hImageList = NULL;
HDC hMemDC = NULL;
HDC hForksDC = NULL;
HBITMAP hMemBmp = NULL;
HBITMAP hForksBmp = NULL;
ColorData colorList[100] = {0, 0, 0};
int colorCount = 0;
#pragma data_seg(pop)
#pragma comment(linker, "/section:SHARED,RWS")

void __declspec(dllexport) __stdcall InitForkOptions();
void __declspec(dllexport) __stdcall SortForks();

const char *GetSelectedName(char *textBuf, int cchTextMax)
{
	HWND hList = GetDlgItem(hDlg, IDC_PF_LIST);
	int item = ListView_GetNextItem(hList, -1, LVNI_SELECTED);
	if(item >= 0)
	{
		LVITEM lvi;
		lvi.iItem = item;
		lvi.iSubItem = 0;
		lvi.mask = LVIF_TEXT;
		lvi.pszText = textBuf;
		lvi.cchTextMax = cchTextMax;
		if(ListView_GetItem(hList, &lvi) == TRUE)
		{
			return lvi.pszText;
		}
	}
	return NULL;
}

void AddCommand(int id, LPCSTR objectName, int param)
{
	if(commandCount < MAX_COMMANDS && strlen(objectName) <= MAX_NAME_LEN)
	{
		commandList[commandCount].id = id;
		strcpy(commandList[commandCount].objectName, objectName);
		commandList[commandCount].param = param;
		commandCount++;
	}
}

int GetMaxLineNum(const char *forkName)
{
	size_t len = strlen(forkName);
	int maxNum = 0;
	for(int i = 0; i < lineCount; i++)
	{
		if(strlen(lineNames[i]) > len && strncmp(forkName, lineNames[i], len) == 0)
		{
			int wl = (atoi(lineNames[i] + len) - 100) / 50;
			if(maxNum < wl)
			{
				maxNum = wl;
			}
		}
	}
	return maxNum;
}

int GetMinLineNum(const char *forkName)
{
	size_t len = strlen(forkName);
	int minNum = 0;
	for(int i = 0; i < lineCount; i++)
	{
		if(strlen(lineNames[i]) > len && strncmp(forkName, lineNames[i], len) == 0)
		{
			int wl = -atoi(lineNames[i] + len) / 50;
			if(minNum < wl)
			{
				minNum = wl;
			}
		}
	}
	return minNum;
}

BOOL CALLBACK DialogProc(HWND hwndDlg,
    UINT uMsg,
    WPARAM wParam,
    LPARAM lParam
)
{
	if(uMsg == WM_INITDIALOG)
	{
		CheckDlgButton(hwndDlg, IDC_SCHIFF, BST_CHECKED);
		SetDlgItemTextA(hwndDlg, IDC_WL_COUNT, "0");
//		The timer refreshes every chart on each tick and increases server load
//		SetTimer(hwndDlg, 0, 5000, NULL);
		SetDlgItemInt(hDlg, IDC_WLA_EXTRA, 0, FALSE);
		SetDlgItemInt(hDlg, IDC_WLB_EXTRA, 0, FALSE);
		SendDlgItemMessage(hwndDlg, IDC_SPIN_WLA_EXTRA, UDM_SETRANGE, 0, MAKELONG((short)19, (short)0));
		SendDlgItemMessage(hwndDlg, IDC_SPIN_WLB_EXTRA, UDM_SETRANGE, 0, MAKELONG((short)19, (short)0));

		return FALSE;
	}
/*	else if(uMsg == WM_TIMER && hMtWnd != NULL)
	{
		timerTicks++;
		if((timerTicks & 1) == 0)
		{
			for(int i = 0; i < hwndCount; i++)
			{
				PostMessageA(hwndList[i], WM_COMMAND, MT4_REFRESH, 0);
			}
		}
		else
		{
	        PostMessageA(hMtWnd, WM_COMMAND, MT4_REFRESH, 0);
		}
	}*/
	else if(uMsg == WM_COMMAND && HIWORD(wParam) == BN_CLICKED)
	{
		UINT id = LOWORD(wParam);
		char textBuf[MAX_NAME_LEN + 1];
		const char *name = GetSelectedName(textBuf, sizeof(textBuf));
		if(name != NULL)
		{
			if(id == IDC_SCHIFF)
			{
				AddCommand(CMD_SCHIFF, name, 0);
			}
			else if(id == IDC_UNSCHIFF)
			{
				AddCommand(CMD_UNSCHIFF, name, 0);
			}
			else if(id == IDC_COPYSCHIFF)
			{
				AddCommand(CMD_COPYSCHIFF, name, 0);
			}
			else if(id == IDC_COPYANDREWS)
			{
				AddCommand(CMD_COPYANDREWS, name, 0);
			}
			else if(id == IDC_PARALLEL)
			{
				AddCommand(CMD_PARALLEL, name, 0);
			}
			else for(int i = 0; controlMap[i][1] >= 0; i++)
			{
				if(id == controlMap[i][1])
				{
					if(IsDlgButtonChecked(hDlg, id) == BST_CHECKED)
					{
						AddCommand(CMD_WL, name, controlMap[i][0]);
					}
					else
					{
						AddCommand(CMD_DELETE_WL, name, controlMap[i][0]);
					}
				}
			}
	        PostMessageA(hMtWnd, WM_COMMAND, MT4_REFRESH, 0);
		}
	}
	else if(uMsg == WM_COMMAND && HIWORD(wParam) == EN_CHANGE)
	{
		char textBuf[MAX_NAME_LEN + 1];
		const char *name = GetSelectedName(textBuf, sizeof(textBuf));
		if(name != NULL)
		{
			UINT id = LOWORD(wParam);
			if(id == IDC_WLA_EXTRA)
			{
				int lastLine = LAST_LINE + GetDlgItemInt(hDlg, IDC_WLA_EXTRA, NULL, FALSE);
				int maxLine = GetMaxLineNum(name);
				if(maxLine < LAST_LINE)
					maxLine = LAST_LINE;
				if(lastLine > maxLine)
				{
					for(int i = maxLine + 1; i <= lastLine; i++)
					{
						AddCommand(CMD_WL, name, 100 + i * 50);
				        PostMessageA(hMtWnd, WM_COMMAND, MT4_REFRESH, 0);
					}
				}
				else if(lastLine < maxLine)
				{
					int end = lastLine < LAST_LINE ? LAST_LINE : lastLine;
					for(int i = maxLine; i > end; i--)
					{
						AddCommand(CMD_DELETE_WL, name, 100 + i * 50);
				        PostMessageA(hMtWnd, WM_COMMAND, MT4_REFRESH, 0);
					}
				}
			}
			else if(id == IDC_WLB_EXTRA)
			{
				int lastLine = LAST_LINE + GetDlgItemInt(hDlg, IDC_WLB_EXTRA, NULL, FALSE);
				int minLine = GetMinLineNum(name);
				if(minLine < LAST_LINE)
					minLine = LAST_LINE;
				if(lastLine > minLine)
				{
					for(int i = minLine + 1; i <= lastLine; i++)
					{
						AddCommand(CMD_WL, name, i * -50);
				        PostMessageA(hMtWnd, WM_COMMAND, MT4_REFRESH, 0);
					}
				}
				else if(lastLine < minLine)
				{
					int end = lastLine < LAST_LINE ? LAST_LINE : lastLine;
					for(int i = minLine; i > end; i--)
					{
						AddCommand(CMD_DELETE_WL, name, i * -50);
				        PostMessageA(hMtWnd, WM_COMMAND, MT4_REFRESH, 0);
					}
				}
			}
		}
	}
	else if(uMsg == WM_NOTIFY)
	{
		LPNMHDR pnmh = (LPNMHDR)lParam;
		if(pnmh->code == NM_CLICK)
		{
			InitForkOptions();
		}
	}
	else if(uMsg == WM_DESTROY)
	{
		// Image list is destroyed with the list view
		hImageList = NULL;
	}
	return FALSE;
}

void CreateSettingsDialog(HWND hWnd)
{
	if(hDlg == NULL)
	{
		hWnd = GetAncestor(hWnd, GA_ROOTOWNER);
		hDlg = CreateDialog(hModule, MAKEINTRESOURCE(IDD_ML_OPTIONS), hWnd, DialogProc);
		if(hDlg != NULL)
		{
			SetFocus(hWnd);
			if(hKey == NULL)
			{
				RegCreateKeyEx(HKEY_CURRENT_USER, registryKey,
					0, NULL, 0,
					KEY_CREATE_SUB_KEY|KEY_QUERY_VALUE|KEY_SET_VALUE,
					NULL, &hKey, NULL);
			}
			if(hKey != NULL)
			{
				RECT rc;
				DWORD type;
				DWORD len = sizeof(rc);
				if(RegQueryValueEx(hKey, windowKey, 0, &type, (BYTE*)(&rc), &len) == ERROR_SUCCESS
					&& type == REG_BINARY && len == sizeof(rc))
				{
					RECT wr;
					GetWindowRect(hDlg, &wr);
					MoveWindow(hDlg, rc.left, rc.top, wr.right - wr.left, wr.bottom - wr.top, TRUE);
				}
				UINT u;
				if(RegQueryValueEx(hKey, snapKey, 0, &type, (BYTE*)&u, &len) == ERROR_SUCCESS)
				{
					CheckDlgButton(hDlg, IDC_SNAP_TO_PIVOT, u);
				}
			}
		}
	}
}

bool FindLine(const char *forkName, int lineNum)
{
	size_t len = strlen(forkName);
	for(int i = 0; i < lineCount; i++)
	{
		if(strlen(lineNames[i]) > len && strncmp(forkName, lineNames[i], len) == 0)
		{
			int wl = atoi(lineNames[i] + len);
			if(wl == lineNum)
			{
				return true;
			}
		}
	}
	return false;
}

int GetWindowIndex(HWND hWnd)
{
	for(int i = 0; i < hwndCount; i++)
	{
		if(hwndList[i] == hWnd)
		{
			return i;
		}
	}
	return -1;
}

void RemoveWindow(HWND hWnd)
{
	int index = GetWindowIndex(hWnd);
	if(index >= 0)
	{
		memmove(hwndList + index, hwndList + index + 1,
			(char*)(hwndList + hwndCount) - (char*)(hwndList + index + 1));
		hwndCount--;
	}
}

int FindImages(COLORREF color)
{
	for(int i = 0; i < colorCount; i++)
	{
		if(colorList[i].color == color)
			return i;
	}
	return -1;
}

int AddImage(int color, int isUp)
{
	if(hImageList == NULL)
	{
		hImageList = ImageList_Create(16, 16, ILC_COLOR24 | ILC_MASK, 20, 20);
		ListView_SetImageList(GetDlgItem(hDlg, IDC_PF_LIST), hImageList, LVSIL_SMALL);
	}
	if(hForksBmp == NULL)
	{
		hForksBmp = LoadBitmap(hModule, MAKEINTRESOURCE(IDB_FORKICONS));
	}
	if(hMemDC == NULL)
	{
		HDC hDC = GetDC(hDlg);
		hMemDC = CreateCompatibleDC(hDC);
		if(hMemBmp == NULL)
		{
			hMemBmp = CreateCompatibleBitmap(hDC, 16, 16);
		}
		hForksDC = CreateCompatibleDC(hDC);
		SelectObject(hForksDC, hForksBmp);
		ReleaseDC(hDlg, hDC);
	}
	int image = FindImages(color);
	if(image < 0)
	{
		if(colorCount < MAX_COLORS)
		{
			RECT rc;
			rc.left = 0;
			rc.top = 0;
			rc.right = 16;
			rc.bottom = 16;
			SetBkColor(hMemDC, color);
			HGDIOBJ oldBmp = SelectObject(hMemDC, hMemBmp);
			ExtTextOutA(hMemDC, 0, 0, ETO_OPAQUE, &rc, NULL, 0, NULL);
			BitBlt(hMemDC, 0, 0, 16, 16, hForksDC, 0, 0, SRCAND);
			SelectObject(hMemDC, oldBmp);
			colorList[colorCount].imageUp = ImageList_AddMasked(hImageList, hMemBmp, 0);
			colorList[colorCount].color = color;
			SelectObject(hMemDC, hMemBmp);
			ExtTextOutA(hMemDC, 0, 0, ETO_OPAQUE, &rc, NULL, 0, NULL);
			BitBlt(hMemDC, 0, 0, 16, 16, hForksDC, 16, 0, SRCAND);
			SelectObject(hMemDC, oldBmp);
			colorList[colorCount].imageDn = ImageList_AddMasked(hImageList, hMemBmp, 0);
			image = colorCount;
			colorCount++;
		}
		else
		{
			image = 0;
		}
	}
	return isUp != 0 ? colorList[image].imageUp
		: colorList[image].imageDn;
}

int FindItem(const char *name)
{
	LVFINDINFO lvfi;
	lvfi.flags = LVFI_STRING;
	lvfi.psz = name;
	return ListView_FindItem(GetDlgItem(hDlg, IDC_PF_LIST), -1, &lvfi);
}

void EnableControls(BOOL enable)
{
	if(IsWindowEnabled(GetDlgItem(hDlg, IDC_SCHIFF)) != enable)
	{
		EnableWindow(GetDlgItem(hDlg, IDC_SCHIFF), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_UNSCHIFF), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_COPYSCHIFF), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_COPYANDREWS), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_PARALLEL), enable);
		for(int i = 0; controlMap[i][1] >= 0; i++)
		{
			EnableWindow(GetDlgItem(hDlg, controlMap[i][1]), enable);
		}
		EnableWindow(GetDlgItem(hDlg, IDC_WLA_EXTRA), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_SPIN_WLA_EXTRA), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_WLB_EXTRA), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_SPIN_WLB_EXTRA), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_EXTRA_WLA), enable);
		EnableWindow(GetDlgItem(hDlg, IDC_EXTRA_WLB), enable);
	}
}

int __declspec(dllexport) __stdcall GetDllVersion()
{
	return dllVersion;
}

void __declspec(dllexport) __stdcall AddWindow(HWND hWnd)
{
	if(hWnd != NULL && hwndCount < MAX_WINDOWS && GetWindowIndex(hWnd) < 0)
	{
		hwndList[hwndCount] = hWnd;
		hwndCount++;
#ifdef _DEBUG
		char msg[80];
		wsprintf(msg, "%d hwnds", hwndCount);
		SetWindowText(hDlg, msg);
#endif
	}
}

void __declspec(dllexport) __stdcall Reset(HWND hWnd, const char *title)
{
	hMtWnd = hWnd;
	if(hDlg != NULL)
	{
		ListView_DeleteAllItems(GetDlgItem(hDlg, IDC_PF_LIST));
		if(title != NULL && title[0] != 0)
		{
			SetWindowText(hDlg, title);
		}
	}
	AddWindow(hWnd);
	lineCount = 0;
}

void __declspec(dllexport) __stdcall AddFork(HWND hWnd, const char *name, int time, int color, int isUp)
{
	CreateSettingsDialog(hWnd);
	if(hDlg != NULL && strlen(name) <= MAX_NAME_LEN)
	{
		strcpy(lastForkName, name);
		LVITEM item;
		item.iItem = 0;
		item.iSubItem = 0;
		item.pszText = lastForkName;
		item.iImage = AddImage(color, isUp);
		item.lParam = time;
		item.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
		ListView_InsertItem(GetDlgItem(hDlg, IDC_PF_LIST), &item);
	}
}

void __declspec(dllexport) __stdcall UpdateFork(const char *name, int time, int color, int isUp)
{
	LVITEM item;
	item.iItem = FindItem(name);
	if(item.iItem >= 0)
	{
		item.iSubItem = 0;
		item.mask = LVIF_IMAGE | LVIF_PARAM;
		if(ListView_GetItem(GetDlgItem(hDlg, IDC_PF_LIST), &item) == TRUE)
		{
			int image = AddImage(color, isUp);
			if(image != item.iImage || time != item.lParam)
			{
				item.iImage = image;
				item.lParam = time;
				ListView_SetItem(GetDlgItem(hDlg, IDC_PF_LIST), &item);
				SortForks();
			}
		}
	}
}

void __declspec(dllexport) __stdcall SelectNewestFork()
{
	int item = FindItem(lastForkName);
	if(item >= 0)
	{
		ListView_SetItemState(GetDlgItem(hDlg, IDC_PF_LIST), item, LVIS_SELECTED, LVIS_SELECTED);
	}
}

void __declspec(dllexport) __stdcall ResetLines()
{
	lineCount = 0;
}

void __declspec(dllexport) __stdcall AddLine(HWND hWnd, const char *name)
{
	if(hDlg != NULL && strlen(name) <= MAX_NAME_LEN && lineCount < MAX_LINES)
	{
		strcpy(lineNames[lineCount], name);
		lineCount++;
	}
}

void __declspec(dllexport) __stdcall InitForkOptions()
{
	char textBuf[MAX_NAME_LEN + 1];
	const char *name = GetSelectedName(textBuf, sizeof(textBuf));
	if(name != NULL)
	{
		EnableControls(TRUE);
		for(int i = 0; controlMap[i][1] >= 0; i++)
		{
			if(FindLine(name, controlMap[i][0]))
				CheckDlgButton(hDlg, controlMap[i][1], BST_CHECKED);
			else CheckDlgButton(hDlg, controlMap[i][1], BST_UNCHECKED);
		}
		int maxLine = GetMaxLineNum(name);
		int currExtra = GetDlgItemInt(hDlg, IDC_WLA_EXTRA, NULL, FALSE);
		if(maxLine > LAST_LINE)
		{
			if(currExtra != maxLine - LAST_LINE)
			{
				SetDlgItemInt(hDlg, IDC_WLA_EXTRA, maxLine - LAST_LINE, FALSE);
			}
		}
		else if(currExtra != 0)
		{
			SetDlgItemInt(hDlg, IDC_WLA_EXTRA, 0, FALSE);
		}
		currExtra = GetDlgItemInt(hDlg, IDC_WLB_EXTRA, NULL, FALSE);
		int minLine = GetMinLineNum(name);
		if(minLine > LAST_LINE)
		{
			if(currExtra != minLine - LAST_LINE)
			{
				SetDlgItemInt(hDlg, IDC_WLB_EXTRA, minLine - LAST_LINE, FALSE);
			}
		}
		else if(currExtra != 0)
		{
			SetDlgItemInt(hDlg, IDC_WLB_EXTRA, 0, FALSE);
		}
	}
	else EnableControls(FALSE);
}

void __declspec(dllexport) __stdcall _DeleteObject(const char *name)
{
	int item = FindItem(name);
	if(item >= 0)
	{
		ListView_DeleteItem(GetDlgItem(hDlg, IDC_PF_LIST), item);
	}
}

int CALLBACK CompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	return lParam2 - lParam1;
}

void __declspec(dllexport) __stdcall SortForks()
{
	ListView_SortItems(GetDlgItem(hDlg, IDC_PF_LIST), CompareFunc, 0);
}

HWND __declspec(dllexport) __stdcall GetCurrentWindow()
{
	return hMtWnd;
}

int __declspec(dllexport) __stdcall GetCommandId(int i)
{
	if(i < commandCount)
	{
		return commandList[i].id;
	}
	return -1;
}

LPCSTR __declspec(dllexport) __stdcall GetObjectName(int i)
{
	if(i < commandCount)
	{
		return commandList[i].objectName;
	}
	return NULL;
}

int __declspec(dllexport) __stdcall GetCommandParam(int i)
{
	if(i < commandCount)
	{
		return commandList[i].param;
	}
	return 0;
}

void __declspec(dllexport) __stdcall ClearCommands()
{
	commandCount = 0;
}

int __declspec(dllexport) __stdcall IsSnapChecked()
{
	return IsDlgButtonChecked(hDlg, IDC_SNAP_TO_PIVOT);
}

void __declspec(dllexport) __stdcall CloseDialog(HWND hWnd)
{
	RemoveWindow(hWnd);
	if(hwndCount == 0 && hDlg != NULL)
	{
		RECT rc;
		GetWindowRect(hDlg, &rc);
		RegSetValueEx(hKey, windowKey, 0, REG_BINARY, (BYTE*)(&rc), sizeof(rc));
		UINT u = IsDlgButtonChecked(hDlg, IDC_SNAP_TO_PIVOT);
		RegSetValueEx(hKey, snapKey, 0, REG_DWORD, (BYTE*)(&u), sizeof(u));
		RegCloseKey(hKey);
		hKey = NULL;
		DestroyWindow(hDlg);
		hDlg = NULL;
		hImageList = NULL;
	}
}

